from .policy import JWTAuthenticationPolicy


def includeme(config):
    config.add_directive(
        'set_jwt_authentication_policy',
        set_jwt_authentication_policy,
        action_wrap=True)


def create_jwt_authentication_policy(config, private_key=None, public_key=None,
        algorithm=None, expiration=None, leeway=None,
        http_header=None, auth_type=None, callback=None, json_encoder=None,
        audience=None,):
    settings = config.get_settings()
    private_key = private_key or settings.get('jwt.private_key')
    audience = audience or settings.get('jwt.audience')
    algorithm = algorithm or settings.get('jwt.algorithm') or 'HS512'
    if not algorithm.startswith('HS'):
            public_key = public_key or settings.get('jwt.public_key')
    else:
            public_key = None
    if expiration is None and 'jwt.expiration' in settings:
        expiration = int(settings.get('jwt.expiration'))
    leeway = int(settings.get('jwt.leeway', 0)) if leeway is None else leeway
    http_header = http_header or settings.get('jwt.http_header') or 'Authorization'
    if http_header.lower() == 'authorization':
            auth_type = auth_type or settings.get('jwt.auth_type') or 'JWT'
    else:
            auth_type = None
    return JWTAuthenticationPolicy(
            private_key=private_key,
            public_key=public_key,
            algorithm=algorithm,
            leeway=leeway,
            expiration=expiration,
            http_header=http_header,
            auth_type=auth_type,
            callback=callback,
            json_encoder=json_encoder,
            audience=audience)


def set_jwt_authentication_policy(config, private_key=None, public_key=None,
        algorithm=None, expiration=None, leeway=None,
        http_header=None, auth_type=None, callback=None, json_encoder=None,
        audience=None,):
    policy = create_jwt_authentication_policy(
            config, private_key, public_key,
            algorithm, expiration, leeway,
            http_header, auth_type, callback, json_encoder, audience)

    def request_create_token(request, principal, expiration=None, audience=None, **claims):
        return policy.create_token(principal, expiration, audience, **claims)

    def request_claims(request):
        return policy.get_claims(request)

    config.set_authentication_policy(policy)
    config.add_request_method(request_create_token, 'create_jwt_token')
    config.add_request_method(request_claims, 'jwt_claims', reify=True)



import datetime
import logging
import warnings
import jwt
from zope.interface import implementer
from pyramid.authentication import CallbackAuthenticationPolicy
from pyramid.interfaces import IAuthenticationPolicy


log = logging.getLogger('pyramid_jwt')
marker = []

@implementer(IAuthenticationPolicy)
class JWTAuthenticationPolicy(CallbackAuthenticationPolicy):
    def __init__(self, private_key, public_key=None, algorithm='HS512',
            leeway=0, expiration=None, default_claims=None,
            http_header='Authorization', auth_type='JWT',
            callback=None, json_encoder=None, audience=None,):
        self.private_key = private_key
        self.public_key = public_key if public_key is not None else private_key
        self.algorithm = algorithm
        self.leeway = leeway
        self.default_claims = default_claims if default_claims else {}
        self.http_header = http_header
        self.auth_type = auth_type
        if expiration:
            if not isinstance(expiration, datetime.timedelta):
                    expiration = datetime.timedelta(seconds=expiration)
            self.expiration = expiration
        else:
            self.expiration = None
        if audience:
            self.audience = audience
        else:
            self.audience = None
        self.callback = callback
        self.json_encoder = json_encoder

    def create_token(self, principal, expiration=None, audience=None, **claims):
        payload = self.default_claims.copy()
        payload.update(claims)
        payload['sub'] = principal
        payload['iat'] = iat = datetime.datetime.utcnow()
        expiration = expiration or self.expiration
        audience = audience or self.audience
        if expiration:
            if not isinstance(expiration, datetime.timedelta):
                    expiration = datetime.timedelta(seconds=expiration)
            payload['exp'] = iat + expiration
        if audience:
            payload['aud'] = audience
        token = jwt.encode(payload, self.private_key, algorithm=self.algorithm, json_encoder=self.json_encoder)
        if not isinstance(token, str):  # Python3 unicode madness
            token = token.decode('ascii')
        return token


    def get_claims(self, request):
        if self.http_header == 'Authorization':
            try:
                if request.authorization is None:
                    return {}
            except ValueError:  # Invalid Authorization header
                return {}
            (auth_type, token) = request.authorization
            if auth_type != self.auth_type:
                return {}
        else:
            token = request.headers.get(self.http_header)
        if not token:
            return {}
        try:
            claims = jwt.decode(token, self.public_key, algorithms=[self.algorithm],
                                leeway=self.leeway, audience=self.audience)
            return claims
        except jwt.InvalidTokenError as e:
            log.warning('Invalid JWT token from %s: %s', request.remote_addr, e)
            return {}


    def unauthenticated_userid(self, request):
        return request.jwt_claims.get('sub')

    def remember(self, request, principal, **kw):
        warnings.warn(
            'JWT tokens need to be returned by an API. Using remember() '
            'has no effect.',
            stacklevel=3)
        return []

    def forget(self, request):
        warnings.warn(
            'JWT tokens are managed by API (users) manually. Using forget() '
            'has no effect.',
            stacklevel=3)
        return []



import pytest
from pyramid.config import Configurator
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.security import Authenticated
from pyramid.security import Allow
from webtest import TestApp


def login_view(request):
    return {'token': request.create_jwt_token(1)}


def secure_view(request):
    return 'OK'


class Root:
    __acl__ = [
        (Allow, Authenticated, ('read',)),
    ]

    def __init__(self, request):
        pass


@pytest.fixture
def app():
    config = Configurator()
    config.set_authorization_policy(ACLAuthorizationPolicy())
    # Enable JWT authentication.
    config.include('pyramid_jwt')
    config.set_root_factory(Root)
    config.set_jwt_authentication_policy('secret', http_header='X-Token')
    config.add_route('login', '/login')
    config.add_view(login_view, route_name='login', renderer='json')
    config.add_route('secure', '/secure')
    config.add_view(secure_view, route_name='secure', renderer='string',
         permission='read')
    app = config.make_wsgi_app()
    return TestApp(app)


def test_secure_view_requires_auth(app):
    app.get('/secure', status=403)


def test_login(app):
    r = app.get('/login')
    token = str(r.json_body['token'])  # Must be str on all Python versions
    r = app.get('/secure', headers={'X-Token': token})
    assert r.unicode_body == 'OK'



# vim: fileencoding=utf-8
import warnings
from webob import Request
from zope.interface.verify import verifyObject
from pyramid.security import forget
from pyramid.security import remember
from pyramid.testing import testConfig
from pyramid.testing import DummyRequest
from pyramid.testing import DummySecurityPolicy
from pyramid.interfaces import IAuthenticationPolicy
from pyramid_jwt.policy import JWTAuthenticationPolicy
import uuid
import pytest
from json.encoder import JSONEncoder
from uuid import UUID


def test_interface():
    verifyObject(IAuthenticationPolicy, JWTAuthenticationPolicy('secret'))


def test_token_most_be_str():
    policy = JWTAuthenticationPolicy('secret')
    token = policy.create_token(15)
    assert isinstance(token, str)


def test_minimal_roundtrip():
    policy = JWTAuthenticationPolicy('secret')
    request = Request.blank('/')
    request.authorization = ('JWT', policy.create_token(15))
    request.jwt_claims = policy.get_claims(request)
    assert policy.unauthenticated_userid(request) == 15


def test_audience_valid():
    policy = JWTAuthenticationPolicy('secret', audience='example.org')
    token = policy.create_token(15, name=u'Jöhn', admin=True, audience='example.org')
    request = Request.blank('/')
    request.authorization = ('JWT', token)
    jwt_claims = policy.get_claims(request)
    assert jwt_claims['aud'] == 'example.org'

def test_audience_invalid():
    policy = JWTAuthenticationPolicy('secret', audience='example.org')
    token = policy.create_token(15, name=u'Jöhn', admin=True, audience='example.com')
    request = Request.blank('/')
    request.authorization = ('JWT', token)
    jwt_claims = policy.get_claims(request)
    assert jwt_claims == {}


def test_algorithm_unsupported():
    policy = JWTAuthenticationPolicy('secret', algorithm='SHA1')
    with pytest.raises(NotImplementedError):
        token = policy.create_token(15, name=u'Jöhn', admin=True )


def test_extra_claims():
    policy = JWTAuthenticationPolicy('secret')
    token = policy.create_token(15, name=u'Jöhn', admin=True)
    request = Request.blank('/')
    request.authorization = ('JWT', token)
    jwt_claims = policy.get_claims(request)
    assert jwt_claims['name'] == u'Jöhn'
    assert jwt_claims['admin']


def test_wrong_auth_scheme():
    policy = JWTAuthenticationPolicy('secret')
    request = Request.blank('/')
    request.authorization = ('Other', policy.create_token(15))
    request.jwt_claims = policy.get_claims(request)
    assert policy.unauthenticated_userid(request) is None


def test_invalid_authorization_header():
    policy = JWTAuthenticationPolicy('secret')
    request = Request.blank('/')
    request.environ['HTTP_AUTHORIZATION'] = 'token'
    request.jwt_claims = policy.get_claims(request)
    assert policy.unauthenticated_userid(request) is None


def test_other_header():
    policy = JWTAuthenticationPolicy('secret', http_header='X-Token')
    request = Request.blank('/')
    request.headers['X-Token'] = policy.create_token(15)
    request.jwt_claims = policy.get_claims(request)
    assert policy.unauthenticated_userid(request) == 15


def test_expired_token():
    policy = JWTAuthenticationPolicy('secret', expiration=-1)
    request = Request.blank('/')
    request.authorization = ('JWT', policy.create_token(15))
    request.jwt_claims = policy.get_claims(request)
    assert policy.unauthenticated_userid(request) is None
    policy.leeway = 5
    request.jwt_claims = policy.get_claims(request)
    assert policy.unauthenticated_userid(request) == 15


def test_dynamic_expired_token():
    policy = JWTAuthenticationPolicy('secret', expiration=-1)
    request = Request.blank('/')
    request.authorization = ('JWT', policy.create_token(15, expiration=5))
    request.jwt_claims = policy.get_claims(request)
    assert policy.unauthenticated_userid(request) == 15

    policy = JWTAuthenticationPolicy('secret')
    request.authorization = ('JWT', policy.create_token(15, expiration=-1))
    request.jwt_claims = policy.get_claims(request)
    assert policy.unauthenticated_userid(request) is None
    request.authorization = ('JWT', policy.create_token(15))
    request.jwt_claims = policy.get_claims(request)
    assert policy.unauthenticated_userid(request) == 15


def test_remember_warning():
    policy = JWTAuthenticationPolicy('secret', http_header='X-Token')
    with testConfig() as config:
        config.set_authorization_policy(DummySecurityPolicy())
        config.set_authentication_policy(policy)
        request = DummyRequest()
        with warnings.catch_warnings(record=True) as w:
            remember(request, 15)
    assert len(w) == 1
    assert issubclass(w[-1].category, UserWarning)
    assert 'JWT tokens' in str(w[-1].message)
    assert w[-1].filename.endswith('test_policy.py')


def test_forget_warning():
    policy = JWTAuthenticationPolicy('secret', http_header='X-Token')
    with testConfig() as config:
        config.set_authorization_policy(DummySecurityPolicy())
        config.set_authentication_policy(policy)
        request = DummyRequest()
        with warnings.catch_warnings(record=True) as w:
            forget(request)
    assert len(w) == 1
    assert issubclass(w[-1].category, UserWarning)
    assert 'JWT tokens' in str(w[-1].message)
    assert w[-1].filename.endswith('test_policy.py')
    
class MyCustomJsonEncoder(JSONEncoder):

    def default(self, o):
        if type(o) is UUID:
            return str(o)
        # Let the base class default method raise the TypeError
        return JSONEncoder.default(self, o)

def test_custom_json_encoder():
    policy = JWTAuthenticationPolicy('secret')
    principal_id = uuid.uuid4()
    claim_value = uuid.uuid4()
    with pytest.raises(TypeError):
        token = policy.create_token('subject', uuid_value=claim_value)
    policy = JWTAuthenticationPolicy('secret', json_encoder=MyCustomJsonEncoder)
    
    request = Request.blank('/')
    request.authorization = ('JWT', policy.create_token(principal_id, uuid_value=claim_value))
    request.jwt_claims = policy.get_claims(request)
    assert policy.unauthenticated_userid(request) == str(principal_id)
    assert request.jwt_claims.get('uuid_value') == str(claim_value)
    



import os
import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand

version = '1.4.1'

install_requires = [
    'pyramid',
    'PyJWT',
]

tests_require = [
    'pytest',
    'WebTest',
]


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['tests']
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


setup(name='pyramid_jwt',
      version=version,
      description='JWT authentication policy for Pyramid',
      long_description=open('README.rst').read() + '\n' +
              open(os.path.join('docs', 'example.rst')).read() + '\n' +
              open('changes.rst').read(),
      classifiers=[
          'Intended Audience :: Developers',
          'License :: DFSG approved',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      keywords='Pyramid JWT authentication security',
      author='Wichert Akkerman',
      author_email='wichert@wiggy.net',
      url='https://github.com/wichert/pyramid_jwt',
      license='BSD',
      packages=['pyramid_jwt'],
      package_dir={'': 'src'},
      include_package_data=True,
      zip_safe=True,
      install_requires=install_requires,
      tests_require=tests_require,
      extras_require={'tests': tests_require},
      cmdclass={'test': PyTest},
      )
